// fbo.cpp
// PRACA S FRAMEBUFFER OBJEKTAMI, RENDEROVANIE DO TEXTUR

#include "fbo.h"

FBO CreateFBO(int width, int height, int num_color_buffers, bool depth_buffer)
{
    // zisti ci sme neprekrocili povoleny pocet color bufrov
    if (num_color_buffers > 4)
        num_color_buffers = 4;
    GLint maxColorAttachments;
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &maxColorAttachments);
    if (num_color_buffers > maxColorAttachments)
        num_color_buffers = maxColorAttachments;

    // vytvor FBO
    FBO fbo;
    glGenFramebuffersEXT(1, &fbo.id);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.id);

    // vytvor dany pocet texture a pripoj ich ku color bufrom
    for (int i = 0; i < num_color_buffers; i++)
    {
        glGenTextures(1, &(fbo.color_textures[i]));
        glBindTexture(GL_TEXTURE_2D, fbo.color_textures[i]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_SHORT, 0);
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + i, GL_TEXTURE_2D, fbo.color_textures[i], 0);
    }

    // pridaj texturu alebo renderbuffer ako depth buffer
    if (depth_buffer)
    {
        glGenTextures(1, &(fbo.depth_texture));
        glBindTexture(GL_TEXTURE_2D, fbo.depth_texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
        //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, width, height, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_SHORT, 0);
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, fbo.depth_texture, 0);
    }
    else
    {
        GLuint rboId;
        glGenRenderbuffersEXT(1, &rboId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rboId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height);
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, rboId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
    }

    // nastav buffre pre zapis
    if (num_color_buffers == 0)
        glDrawBuffer(GL_NONE);
    else if (num_color_buffers == 1)
        glDrawBuffer(GL_COLOR_ATTACHMENT0);
    else
    {
        GLuint attachments[4] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1, GL_COLOR_ATTACHMENT2, GL_COLOR_ATTACHMENT3 };
        glDrawBuffers(num_color_buffers, attachments);
    }

    // zisti ci takto nastaveny framebuffer je OK
    GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
    {
        printf("FBO Creation Error!\n");
        fbo.id = 0;
    }
    else
        printf("FBO was created successfully...\n");

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    return fbo;
}

void SetMirrorMatrices(aiVector3D camera_pos, std::vector<aiVector3D> mirror_points)
{
    int num_points = (int)mirror_points.size();
    if (num_points < 3)
        return;

    // najdi analyticke vyjadrenie roviny zrkadla na zaklade n bodov
    // majdi zaroven tazisko n-uholnika
    float nx = 0, ny = 0, nz = 0;
    float cx = 0, cy = 0, cz = 0;
    for (int i = 0; i < num_points; i++) 
    {
        int j = (i + 1) % num_points;
        nx += (mirror_points[i].y - mirror_points[j].y) * (mirror_points[i].z + mirror_points[j].z);
        ny += (mirror_points[i].z - mirror_points[j].z) * (mirror_points[i].x + mirror_points[j].x);
        nz += (mirror_points[i].x - mirror_points[j].x) * (mirror_points[i].y + mirror_points[j].y);
        cx += mirror_points[i].x;
        cy += mirror_points[i].y;
        cz += mirror_points[i].z;
    }
    cx = cx / float(num_points);
    cy = cy / float(num_points);
    cz = cz / float(num_points);
    float d = -(nx * mirror_points[0].x + ny * mirror_points[0].y + nz * mirror_points[0].z);
    // nx, ny, nz, d su koeficienty analytickeho vyjadrenia roviny zrkadla

    // zrkadli poziciu kamery vzhladom na zrkadlo
    float t = -(nx * camera_pos.x + ny * camera_pos.y + nz * camera_pos.z + d) / (nx * nx + ny * ny + nz * nz);
    float mx = camera_pos.x + 2 * t * nx;
    float my = camera_pos.y + 2 * t * ny;
    float mz = camera_pos.z + 2 * t * nz;
    float dist = abs(t) * sqrt(nx * nx + ny * ny + nz * nz);

    // zadaj modelview maticu kamery za zrkadlom
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    gluLookAt(mx, my, mz, camera_pos.x, camera_pos.y, camera_pos.z, 0, 1, 1);
    GLdouble mirror_view_matrix[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, mirror_view_matrix);

    // pre kazdy mirror vrchol, strasformuj ho do kamerovych suradnic a najdi extremne hodnoty
    GLdouble bottom = 9999999, top = -9999999, left = 9999999, right = -9999999;
    for (int i = 0; i < num_points; i++)
    {
        GLdouble tx = mirror_view_matrix[0]*mirror_points[i].x+mirror_view_matrix[4]*mirror_points[i].y+mirror_view_matrix[8]*mirror_points[i].z+mirror_view_matrix[12]*1;
        GLdouble ty = mirror_view_matrix[1]*mirror_points[i].x+mirror_view_matrix[5]*mirror_points[i].y+mirror_view_matrix[9]*mirror_points[i].z+mirror_view_matrix[13]*1;
        if (tx < left) left = tx;
        if (tx > right) right = tx;
        if (ty < bottom) bottom = ty;
        if (ty > top) top = ty;
    }
    // zadaj projekcnu maticu
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(left, right, bottom, top, dist, 600);

    glMatrixMode(GL_MODELVIEW);
}

void RenderMirror(std::vector<aiVector3D> mirror_points, GLdouble mirror_view_matrix[16], GLdouble mirror_projection_matrix[16], GLuint mirror_tex)
{
    glUseProgram(0);
    glBindTexture(GL_TEXTURE_2D, mirror_tex);
    glEnable(GL_TEXTURE_2D);

    glBegin(GL_POLYGON);
    GLint viewport[4] = {0, 0, 1024, 1024};
    for (int i = 0; i < (int)mirror_points.size(); i++)
    {
        // vypocitaj projekciu vrchola zrkadla v priemetni, tym sa urcia texturne suradnice
        GLdouble wx, wy, wz;
        gluProject(mirror_points[i].x, mirror_points[i].y, mirror_points[i].z, mirror_view_matrix, mirror_projection_matrix, viewport, &wx, &wy, &wz);
        glTexCoord2d(wx / 1024.0, wy / 1024.0);
        glVertex3f(mirror_points[i].x, mirror_points[i].y, mirror_points[i].z);
    }
    glEnd();
}